/*
 * Linux initialization.
 * Copyright (C) 1995 Shantanu Goel.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  linux/init/main.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#define LINUX_MACH_INCLUDE

#include <sys/types.h>

#include <mach/machine/vm_param.h>
#include <mach/vm_param.h>
#include <mach/vm_prot.h>
#include <mach/machine.h>

#include <vm/vm_page.h>

#include <i386/ipl.h>
#include <i386/pic.h>
#include <i386/pit.h>
#include <i386/machspl.h>
#include <i386/pmap.h>

#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/mm.h>

#include <asm/system.h>

/*
 * Set if the machine has an EISA bus.
 */
int EISA_bus = 0;

/*
 * Timing loop count.
 */
unsigned long linux_loops_per_sec = 1;

/*
 * End of physical memory.
 */
unsigned long linux_high_memory;

/*
 * Flag to indication auto-configuration is in progress.
 */
int linux_auto_config = 1;

/*
 * Forward declarations.
 */
static void alloc_contig_mem(vm_page_t *, unsigned *, unsigned *);
static void free_contig_mem(vm_page_t, unsigned, unsigned);
static void calibrate_delay(void);
static int linux_clock();

extern int hz;
extern volatile unsigned long elapsed_ticks;
extern vm_offset_t phys_last_addr;

extern void linux_timer_soft(void *);
extern void startrtclock(void);
extern unsigned linux_kmem_init(unsigned, unsigned);
extern void linux_net_dev_init(void);
extern void linux_printk(char *, ...);

/*
 * Uncomment this if you want memory usage messages for debugging.
 * It also enables a 4 second delay after device probe.
 */
#define INIT_DEBUG

/*
 * Initialize Linux drivers.
 */
void
linux_init()
{
	char *p;
	int i;
	int (*old_clock_handler)(), old_clock_pri;
	unsigned memory_start, memory_end;
#ifdef INIT_DEBUG
	unsigned saved_memory_start;
#endif
	vm_page_t pages;

	/*
	 * Initialize memory size.
	 */
	linux_high_memory = phys_last_addr;

	/*
	 * Ensure interrupts are disabled.
	 */
	(void) splhigh();

	/*
	 * Program counter 0 of 8253 to interrupt hz times per second.
	 */
	outb(PITCTL_PORT, PIT_C0|PIT_SQUAREMODE|PIT_READMODE);
	outb(PITCTR0_PORT, CLKNUM / hz);
	outb(PITCTR0_PORT, (CLKNUM / hz) >> 8);

	/*
	 * Install our clock interrupt handler.
	 */
	old_clock_handler = ivect[0];
	old_clock_pri = intpri[0];
	ivect[0] = linux_clock;
	intpri[0] = SPLHI;
	form_pic_mask();

	/*
	 * Enable interrupts.
	 */
	(void) spl0();

	/*
	 * Check if the machine has an EISA bus.
	 */
	p = (char *)0x0FFFD9;
	if (*p++ == 'E' && *p++ == 'I' && *p++ == 'S' && *p == 'A')
		EISA_bus = 1;

	/*
	 * Permanently allocate standard device ports.
	 */
	linux_request_region(0x00, 0x20, "dma1");
	linux_request_region(0x40, 0x20, "timer");
	linux_request_region(0x70, 0x10, "rtc");
	linux_request_region(0x80, 0x20, "dma page reg");
	linux_request_region(0xc0, 0x20, "dma2");
	linux_request_region(0xf0, 0x02, "fpu");
	linux_request_region(0xf8, 0x08, "fpu");

	/*
	 * Install timer software interrupt handler.
	 */
	linux_bh_base[TIMER_BH].routine = linux_timer_soft;
	linux_bh_base[TIMER_BH].data = 0;
	enable_bh(TIMER_BH);

	/*
	 * Set loop count.
	 */
	calibrate_delay();

	/*
	 * Allocate contiguous memory below 16 MB.
	 */
	alloc_contig_mem(&pages, &memory_start, &memory_end);

#ifdef INIT_DEBUG
	saved_memory_start = memory_start;
	printf("Total memory allocated = %uK\n",
	       (memory_end - memory_start) >> 10);
#endif

	/*
	 * Initialize Linux memory allocator.
	 */
	memory_start = linux_kmem_init(memory_start, memory_end);

	if (memory_start > memory_end)
		panic("linux_init: ran out memory");

	/*
	 * Free unused memory.
	 */
	free_contig_mem(pages, memory_start, memory_end);

	/*
	 * Initialize network devices.
	 */
	linux_net_dev_init();

#ifdef INIT_DEBUG
	printf("Memory actually used = %uK\n",
	       (memory_start - saved_memory_start) >> 10);

	__delay(4 * linux_loops_per_sec);
#endif

	/*
	 * Disable interrupts.
	 */
	(void) splhigh();

	/*
	 * Restore clock interrupt handler.
	 */
	ivect[0] = old_clock_handler;
	intpri[0] = old_clock_pri;
	form_pic_mask();

	linux_auto_config = 0;
}

/*
 * Maximum amount of memory to allocate.
 */
#define MAX_MEM	((2 * 1024 * 1024) / PAGE_SIZE)

/*
 * Maximum DMA addressable memory.
 */
#define DMA_MAX	((16 * 1024 * 1024) / PAGE_SIZE)

#ifndef NBBY
#define NBBY	8
#endif
#ifndef NBPW
#define NBPW	(NBBY * sizeof(int))
#endif

/*
 * Return a list of contiguous physical pages below 16 MB.
 */
static void
alloc_contig_mem(pages, memory_start, memory_end)
	vm_page_t *pages;
	unsigned *memory_start;
	unsigned *memory_end;
{
	int *bits, i, j, n;
	int npages, max, count, bit, mask, maxmem;
	int first_page, cur_page, last_page;
	vm_offset_t addr;
	vm_page_t p, prevp, nextp, page_list, pp, np;

	/*
	 * Compute how much memory to allocate.
	 * Don't use more than 1/2 of total memory.
	 */
	if (MAX_MEM > vm_page_free_count / 2)
		maxmem = vm_page_free_count / 2;
	else
		maxmem = MAX_MEM;

	/*
	 * Allocate bit array.
	 */
	n = (DMA_MAX + NBPW - 1) / NBPW;
	i = n * NBPW;
	bits = (unsigned *)kalloc(i);
	if (bits == 0)
		panic("alloc_contig_mem: unable alloc bit array\n");
	bzero((char *)bits, i);

	/*
	 * Walk the page free list and set a bit for
	 * every usable page in bit array.
	 */
	simple_lock(&vm_page_queue_free_lock);
	for (p = vm_page_queue_free; p; p = (vm_page_t)p->pageq.next) {
		i = p->phys_addr / PAGE_SIZE;
		if (i < DMA_MAX)
			bits[i / NBPW] |= 1 << (i % NBPW);
	}

	/*
	 * Search for contiguous pages by scanning bit array.
	 */
	for (i = max = npages = 0, cur_page = -1; i < n; i++) {
		for (bit = 1, j = 0; j < NBPW; j++, bit <<= 1) {
			if (bits[i] & bit) {
				if (npages == 0) {
					npages = 1;
					cur_page = i * NBPW + j;
				} else if (++npages == maxmem) {
					first_page = cur_page;
					goto done;
				}
			} else if (npages) {
				if (npages > max) {
					max = npages;
					first_page = cur_page;
				}
				cur_page = -1;
				npages = 0;
			}
		}
	}
	if (npages < max)
		npages = max;
	else
		first_page = cur_page;

 done:
	/*
	 * Remove pages from the free list,
	 * and build page list to return to caller.
	 */
	vm_page_free_count -= npages;
	p = vm_page_queue_free;
	prevp = 0;
	page_list = 0;
	last_page = first_page + npages;
	while (1) {
		i = p->phys_addr / PAGE_SIZE;
		if (i >= first_page && i < last_page) {
			nextp = (vm_page_t)p->pageq.next;
			if (prevp)
				prevp->pageq.next = (queue_entry_t)nextp;
			else
				vm_page_queue_free = nextp;
			p->free = FALSE;
			np = page_list;
			pp = 0;
			while (np) {
				if (np->phys_addr > p->phys_addr) {
					p->pageq.next = (queue_entry_t)np;
					if (pp)
						(pp->pageq.next
						 = (queue_entry_t)p);
					else
						page_list = p;
					break;
				}
				pp = np;
				np = (vm_page_t)np->pageq.next;
			}
			if (np == 0) {
				p->pageq.next = 0;
				if (pp)
					pp->pageq.next = (queue_entry_t)p;
				else
					page_list = p;
			}
			if (--npages == 0)
				break;
			p = nextp;
		} else {
			prevp = p;
			p = (vm_page_t)p->pageq.next;
		}
	}

	simple_unlock(&vm_page_queue_free_lock);
	kfree((vm_offset_t)bits, n * NBPW);

	*memory_start = first_page * PAGE_SIZE;
	*memory_end = last_page * PAGE_SIZE;
	*pages = page_list;
}

/*
 * Free contiguous memory.
 */
void
free_contig_mem(pages, memory_start, memory_end)
	vm_page_t pages;
	unsigned memory_start;
	unsigned memory_end;
{
	vm_page_t p, pp;

	memory_start = round_page(memory_start);
	if (memory_start == memory_end)
		return;

	/*
	 * Skip in-use pages.
	 */
	for (p = pages;
	     p->phys_addr < memory_start;
	     p = (vm_page_t)p->pageq.next)
		;

	/*
	 * Mark unused pages as free.
	 */
	for (pp = p;; pp = (vm_page_t)pp->pageq.next) {
		pp->free = TRUE;
		if (!pp->pageq.next)
			break;
	}

	/*
	 * Put pages on free list.
	 */
	simple_lock(&vm_page_queue_free_lock);
	pp->pageq.next = (queue_entry_t)vm_page_queue_free;
	vm_page_queue_free = p;
	vm_page_free_count += (memory_end - memory_start) / PAGE_SIZE;
	simple_unlock(&vm_page_queue_free_lock);
}

/*
 * Calibrate delay loop.
 * Lifted straight from Linux.
 */
static void
calibrate_delay()
{
	int ticks;

	linux_printk("Calibrating delay loop.. ");
	while (linux_loops_per_sec <<= 1) {
		/* Wait for "start of" clock tick.  */
		ticks = elapsed_ticks;
		while (ticks == elapsed_ticks)
			/* nothing */;
		/* Go .. */
		ticks = elapsed_ticks;
		__delay(linux_loops_per_sec);
		ticks = elapsed_ticks - ticks;
		if (ticks >= hz) {
			linux_loops_per_sec = muldiv(linux_loops_per_sec,
						     hz, ticks);
			linux_printk("ok - %lu.%02lu BogoMips\n",
				     linux_loops_per_sec / 500000,
				     (linux_loops_per_sec / 5000) % 100);
			return;
		}
	}
	linux_printk("failed\n");
}

/*
 * Clock interrupt handler.
 */
int
linux_clock()
{
	elapsed_ticks++;
}
