/*
 * Linux IRQ management.
 * Copyright (C) 1995 Shantanu Goel.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *	linux/arch/i386/kernel/irq.c
 *
 *	Copyright (C) 1992 Linus Torvalds
 */

#define LINUX_MACH_INCLUDE

#include <sys/types.h>
#include <i386/machspl.h>
#include <i386/ipl.h>
#include <i386/pic.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/bitops.h>

/*
 * Priority at which a Linux handler should be called.
 * This is used at the time of an IRQ allocation.  It is
 * set by emulation routines for each class of device.
 */
spl_t linux_intr_pri;

/*
 * Flag indicating an interrupt is being handled.
 */
unsigned long linux_intr_count = 0;

/*
 * List of Linux interrupt handlers.
 */
static int (*linux_handlers[16])();

extern spl_t curr_ipl;
extern int curr_pic_mask;
extern int pic_mask[];
extern volatile unsigned long elapsed_ticks;

extern int intnull(), prtnull();

/*
 * Generic interrupt handler for Linux devices.
 * Set up a fake `struct pt_regs' then call the real handler.
 */
static int
linux_intr(irq)
	int irq;
{
	struct pt_regs regs;

	linux_intr_count++;
	(*linux_handlers[irq])(irq, &regs);
	linux_intr_count--;
}

/*
 * Mask an IRQ.
 */
void
linux_disable_irq(irq)
	int irq;
{
	int i, flags;

	save_flags(flags);
	cli();
	for (i = 0; i < intpri[irq]; i++)
		pic_mask[i] |= 1 << irq;
	if (curr_pic_mask != pic_mask[curr_ipl]) {
		curr_pic_mask = pic_mask[curr_ipl];
		outb(PIC_MASTER_OCW, curr_pic_mask);
		outb(PIC_SLAVE_OCW, curr_pic_mask >> 8);
	}
	restore_flags(flags);
}

/*
 * Unmask an IRQ.
 */
void
linux_enable_irq(irq)
	int irq;
{
	int mask, i, flags;

	mask = 1 << irq;
	if (irq >= 8)
		mask |= 1 << 2;
	save_flags(flags);
	cli();
	for (i = 0; i < intpri[irq]; i++)
		pic_mask[i] &= ~mask;
	if (curr_pic_mask != pic_mask[curr_ipl]) {
		curr_pic_mask = pic_mask[curr_ipl];
		outb(PIC_MASTER_OCW, curr_pic_mask);
		outb(PIC_SLAVE_OCW, curr_pic_mask >> 8);
	}
	restore_flags(flags);
}

/*
 * Attach a handler to an IRQ.
 */
int
linux_request_irq(irq, handler, flags, name)
	unsigned int irq;
	int (*handler)();
	unsigned long flags;
	char *name;
{
	if (irq > 15)
		panic("linux_request_irq: bad irq number");

	if (ivect[irq] == intnull || ivect[irq] == prtnull) {
		if (!handler)
			return (-LINUX_EINVAL);
		linux_handlers[irq] = handler;
		ivect[irq] = linux_intr;
		iunit[irq] = irq;
		intpri[irq] = linux_intr_pri;
		linux_enable_irq(irq);
		return (0);
	}
	return (-LINUX_EBUSY);
}

/*
 * Deallocate an irq.
 */
void
linux_free_irq(irq)
	int irq;
{
	if (irq < 0 || irq > 15)
		panic("linux_free_irq: bad irq number");

	linux_disable_irq(irq);
	ivect[irq] = (irq == 7) ? prtnull : intnull;
	iunit[irq] = irq;
	intpri[irq] = SPL0;
}

/*
 * IRQ probe interrupt handler.
 */
void
probe_intr(irq)
	int irq;
{
	linux_disable_irq(irq);
}

/*
 * Set for an irq probe.
 */
unsigned int
linux_probe_irq_on()
{
	unsigned i, irqs = 0;
	unsigned long delay;
	static unsigned fixed_irqs = 0x2147;

	/*
	 * Allocate all available IRQs.
	 */
	for (i = 15; i > 0; i--)
		if (test_bit(1, (void *)&fixed_irqs) == 0
		    && linux_request_irq(i, probe_intr, 0, "probe") == 0)
			irqs |= 1 << i;

	/*
	 * Wait for spurious interrupts to mask themselves out.
	 */
	for (delay = elapsed_ticks + 2; delay > elapsed_ticks; )
		;

	/*
	 * Free IRQs that caused spurious interrupts.
	 */
	for (i = 15; i > 0; i--) {
		if (irqs & (1 << i) & pic_mask[SPL0]) {
			linux_free_irq(i);
			irqs ^= 1 << i;
		}
	}

	return (irqs);
}

/*
 * Return the result of an irq probe.
 */
int
linux_probe_irq_off(irqs)
	unsigned irqs;
{
	unsigned i;

	/*
	 * Deallocate IRQs.
	 */
	for (i = 15; i > 0; i--)
		if (irqs & (1 << i))
			linux_free_irq(i);

	/*
	 * Return IRQ number.
	 */
	irqs &= pic_mask[SPL0];
	if (!irqs)
		return (0);
	i = ffz(~irqs);
	if (irqs != (irqs & (1 << i)))
		i = -i;
	return (i);
}
