/*
 * Linux memory allocation.
 * Copyright (C) 1995 Shantanu Goel.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * XXX: Needs to be dynamic.
 */

#define LINUX_MACH_INCLUDE

#include <sys/types.h>
#include <kern/zalloc.h>
#include <vm/vm_kern.h>
#include <vm/vm_object.h>
#include <vm/vm_map.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <asm/system.h>

/*
 * This structure describes block sizes.
 */
static struct block {
	unsigned size;	/* size of block */
	int pages;	/* # pages to reserve */
	zone_t zone;	/* zone of allocation */
} blocks[] = {
	{   32,  1, 0 },
	{   64,  1, 0 },
	{  128,  1, 0 },
	{  256,  1, 0 },
	{  512,  1, 0 },
	{ 1024,  1, 0 },
	{ 2048,  1, 0 },
	{ 4096,  2, 0 },
};

#define NBLOCKS	(sizeof(blocks) / sizeof(blocks[0]))

static unsigned mem_start, mem_end;

/*
 * Initialize memory allocator.
 */
unsigned
linux_kmem_init(memory_start, memory_end)
	unsigned memory_start;
	unsigned memory_end;
{
	int i;

	mem_start = memory_start = round_page(memory_start);

	/*
	 * Initialize block zones.
	 */
	for (i = 0; i < NBLOCKS; i++) {
		blocks[i].zone = zinit(blocks[i].size,
				       blocks[i].pages * PAGE_SIZE,
				       PAGE_SIZE, FALSE, "");
		zcram(blocks[i].zone, memory_start,
		      blocks[i].pages * PAGE_SIZE);
		memory_start += blocks[i].pages * PAGE_SIZE;
	}

	mem_end = memory_start;

	return (memory_start);
}

/*
 * Find the first block that is >= size.
 */
static inline struct block *
find_block(size)
	unsigned size;
{
	int i;

	for (i = 0; i < NBLOCKS; i++)
		if (blocks[i].size >= size)
			return (blocks + i);
	return (0);
}

/*
 * Allocate memory.
 */
void *
linux_kmalloc(size, priority)
	unsigned int size;
	int priority;
{
	int flags;
	void *mem;
	struct block *b;

	/*
	 * Find block of proper size.
	 */
	b = find_block(size);
	if (!b) {
		extern int linux_auto_config;

		if ((priority & GFP_DMA) == 0 || !linux_auto_config) {
			printf("Size %u too big\n", size);
			panic("linux_kmalloc");
		} else {
			/*
			 * Hack for allocating big chunks of DMA memory
			 * during auto-configuration.
			 */
			return ((void *)alloc_dma_mem(size, 64*1024));
		}
	}

	/*
	 * Allocate memory.
	 */
	save_flags(flags);
	cli();
	mem = (void *)zget(b->zone);
	restore_flags(flags);
	if (mem == 0)
		printf("linux_kmalloc: ran out of %u byte blocks\n", b->size);
	return (mem);
}

/*
 * Free memory.
 */
void
linux_kfree_s(mem, size)
	void *mem;
	int size;
{
	int flags;
	struct block *b;

	if ((unsigned)mem < mem_start || (unsigned)mem >= mem_end)
		return;

	b = find_block(size);
	if (!b)
		panic("linux_kfree_s: block not found");

	save_flags(flags);
	cli();
	zfree(b->zone, (vm_offset_t)mem);
	restore_flags(flags);
}
